/*
DO NOT EDIT!

This file is autogenerated by astropy/wcs/setup_package.py.  To edit
its contents, edit astropy/wcs/docstrings.py
*/

#ifndef __DOCSTRINGS_H__
#define __DOCSTRINGS_H__

extern char doc_Auxprm[198];
extern char doc_Celprm[302];
extern char doc_DistortionLookupTable[371];
extern char doc_InconsistentAxisTypes[100];
extern char doc_InvalidCoordinate[76];
extern char doc_InvalidPrjParameters[75];
extern char doc_InvalidSubimageSpecification[77];
extern char doc_InvalidTabularParameters[76];
extern char doc_InvalidTransform[107];
extern char doc_K[202];
extern char doc_M[56];
extern char doc_NoSolution[68];
extern char doc_NoWcsKeywordsFound[76];
extern char doc_NonseparableSubimageCoordinateSystem[88];
extern char doc_Prjprm[318];
extern char doc_SingularMatrix[70];
extern char doc_Sip[1115];
extern char doc_Tabprm[234];
extern char doc_Wcs[463];
extern char doc_WcsError[39];
extern char doc_Wcsprm[3022];
extern char doc_Wtbarr[202];
extern char doc_a[264];
extern char doc_a_order[60];
extern char doc_all_pix2world[1305];
extern char doc_alt[236];
extern char doc_ap[276];
extern char doc_ap_order[61];
extern char doc_aux[87];
extern char doc_axis_types[916];
extern char doc_b[273];
extern char doc_b_order[60];
extern char doc_bepoch[119];
extern char doc_bounds_check[469];
extern char doc_bp[276];
extern char doc_bp_order[61];
extern char doc_cd[1059];
extern char doc_cdelt[308];
extern char doc_cdfix[480];
extern char doc_cel[80];
extern char doc_cel_offset[176];
extern char doc_celfix[171];
extern char doc_celprm_euler[291];
extern char doc_celprm_isolat[286];
extern char doc_celprm_latpreq[345];
extern char doc_celprm_phi0[280];
extern char doc_celprm_prj[171];
extern char doc_celprm_ref[2132];
extern char doc_celprm_theta0[280];
extern char doc_cname[76];
extern char doc_colax[91];
extern char doc_colnum[290];
extern char doc_compare[1412];
extern char doc_convert[121];
extern char doc_coord[277];
extern char doc_copy[40];
extern char doc_cpdis1[106];
extern char doc_cpdis2[106];
extern char doc_cperi[101];
extern char doc_crder[123];
extern char doc_crln_obs[106];
extern char doc_crota[1055];
extern char doc_crpix[88];
extern char doc_crval[93];
extern char doc_crval_tabprm[94];
extern char doc_csyer[132];
extern char doc_ctype[228];
extern char doc_cubeface[1236];
extern char doc_cunit[1323];
extern char doc_cylfix[162];
extern char doc_czphs[125];
extern char doc_data[78];
extern char doc_data_wtbarr[52];
extern char doc_dateavg[152];
extern char doc_datebeg[137];
extern char doc_dateend[135];
extern char doc_dateobs[133];
extern char doc_dateref[133];
extern char doc_datfix[512];
extern char doc_delta[238];
extern char doc_det2im[58];
extern char doc_det2im1[103];
extern char doc_det2im2[103];
extern char doc_dims[98];
extern char doc_dsun_obs[110];
extern char doc_equinox[243];
extern char doc_extlev[72];
extern char doc_extnam[73];
extern char doc_extrema[439];
extern char doc_extver[72];
extern char doc_find_all_wcs[866];
extern char doc_fix[1679];
extern char doc_get_cdelt[474];
extern char doc_get_offset[201];
extern char doc_get_pc[413];
extern char doc_get_ps[431];
extern char doc_get_pv[741];
extern char doc_has_cd[911];
extern char doc_has_cdi_ja[103];
extern char doc_has_crota[652];
extern char doc_has_crotaia[107];
extern char doc_has_pc[213];
extern char doc_has_pci_ja[103];
extern char doc_hgln_obs[100];
extern char doc_hglt_obs[122];
extern char doc_i[40];
extern char doc_imgpix_matrix[173];
extern char doc_is_unity[109];
extern char doc_jepoch[116];
extern char doc_kind[157];
extern char doc_lat[86];
extern char doc_latpole[75];
extern char doc_lattyp[236];
extern char doc_lng[87];
extern char doc_lngtyp[237];
extern char doc_lonpole[77];
extern char doc_m[63];
extern char doc_map[967];
extern char doc_mix[4851];
extern char doc_mjdavg[181];
extern char doc_mjdbeg[181];
extern char doc_mjdend[181];
extern char doc_mjdobs[181];
extern char doc_mjdref[182];
extern char doc_name[74];
extern char doc_naxis[761];
extern char doc_nc[172];
extern char doc_ndim[70];
extern char doc_obsgeo[187];
extern char doc_obsorbit[155];
extern char doc_p0[244];
extern char doc_p2s[2113];
extern char doc_p4_pix2foc[682];
extern char doc_pc[1126];
extern char doc_phi0[289];
extern char doc_pix2foc[667];
extern char doc_piximg_matrix[128];
extern char doc_plephem[126];
extern char doc_print_contents[210];
extern char doc_print_contents_tabprm[211];
extern char doc_print_contents_wtbarr[210];
extern char doc_prjprm_bounds[793];
extern char doc_prjprm_category[225];
extern char doc_prjprm_code[60];
extern char doc_prjprm_conformal[51];
extern char doc_prjprm_divergent[59];
extern char doc_prjprm_equiareal[52];
extern char doc_prjprm_global_projection[105];
extern char doc_prjprm_m[94];
extern char doc_prjprm_n[94];
extern char doc_prjprm_name[43];
extern char doc_prjprm_phi0[197];
extern char doc_prjprm_prjs2x[815];
extern char doc_prjprm_prjx2s[719];
extern char doc_prjprm_pv[921];
extern char doc_prjprm_pvi[706];
extern char doc_prjprm_pvrange[207];
extern char doc_prjprm_r0[191];
extern char doc_prjprm_set[658];
extern char doc_prjprm_simplezen[82];
extern char doc_prjprm_theta0[198];
extern char doc_prjprm_w[304];
extern char doc_prjprm_x0[101];
extern char doc_prjprm_y0[101];
extern char doc_radesys[77];
extern char doc_restfrq[94];
extern char doc_restwav[94];
extern char doc_row[39];
extern char doc_rsun_ref[114];
extern char doc_s2p[1946];
extern char doc_sense[232];
extern char doc_set[1206];
extern char doc_set_celprm[298];
extern char doc_set_ps[413];
extern char doc_set_pv[408];
extern char doc_set_tabprm[330];
extern char doc_sip[84];
extern char doc_sip_foc2pix[683];
extern char doc_sip_pix2foc[683];
extern char doc_spcfix[272];
extern char doc_spec[68];
extern char doc_specsys[145];
extern char doc_sptr[1370];
extern char doc_ssysobs[253];
extern char doc_ssyssrc[150];
extern char doc_sub[3541];
extern char doc_tab[112];
extern char doc_telapse[141];
extern char doc_theta0[289];
extern char doc_timedel[93];
extern char doc_timeoffs[186];
extern char doc_timepixr[151];
extern char doc_timesys[230];
extern char doc_timeunit[264];
extern char doc_timrder[132];
extern char doc_timsyer[119];
extern char doc_to_header[2232];
extern char doc_trefdir[118];
extern char doc_trefpos[110];
extern char doc_tstart[150];
extern char doc_tstop[151];
extern char doc_ttype[107];
extern char doc_unitfix[1047];
extern char doc_velangl[186];
extern char doc_velosys[316];
extern char doc_velref[55];
extern char doc_wcs[83];
extern char doc_wtb[82];
extern char doc_xposure[105];
extern char doc_zsource[98];

#endif

